package com.ijchart.xychart.render.xy ;

import java.awt.BasicStroke ;
import java.awt.Color ;
import java.awt.Paint ;
import java.awt.Shape ;
import java.awt.Stroke ;
import java.awt.geom.Rectangle2D ;
import java.io.Serializable ;

import com.ijchart.xychart.axis.value.AbstractValueAxis ;
import com.ijchart.xychart.chartinfo.RenderShapeCollection ;
import com.ijchart.xychart.constant.ChartOrientation ;
import com.ijchart.xychart.data.IChartData ;
import com.ijchart.xychart.data.XYZNumberChartData ;
import com.ijchart.xychart.dataset.IMatrixDataset ;
import com.ijchart.xychart.render.paint.IPaintScale ;
import com.ijchart.xychart.render.paint.LookupPaintScale ;
import com.ijchart.xychart.shape.IChartShape ;
import com.ijchart.xychart.shape.LineAndShapeChartShape ;
import com.ijchart.xychart.shape.LineChartShape ;
import com.ijchart.xychart.shape.PaintScaleLegendChartShape ;
import com.ijchart.xychart.ui.RectangleEdge ;
import com.ijchart.xychart.util.EqualsUtil ;
import com.ijchart.xychart.util.HashcodeUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>ݼÿ{@link com.ijchart.xychart.data.XYZNumberChartData},(x,y)ֵȷλ,
 * zֵȷɫ.</zh>
 * <en>A render that represents data from a {@link com.ijchart.xychart.data.XYZNumberChartData}
 *  by drawing a color block at each (x, y) point, where the color is a function of the
 * z-value from the dataset.</en>
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class XYShapeRender extends AbstractXYRender implements Cloneable, Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = 8622390584049679451L ;

  /**
   * <zh>ȱʡ{@link IPaintScale}Ķ.</zh>
   * <en>The default paint scale.</en>
   */
  public static final IPaintScale DEFAULT_PAINT_SCALE = new LookupPaintScale() ;

  /**
   * <zh>ȱʡͼαǩ.</zh>
   * <en>The legend label.</en>
   */
  public static final String DEFAULT_LEGEND_LABEL = "Legend" ;

  /**
   * <zh>ȱʡǷɼ.</zh>
   * <en>The default flag indicating whether or not guide lines should be drawn for every item.</en>
   */
  public static final boolean DEFAULT_GUIDE_LINE_VISIBLE = false ;

  /**
   * <zh>ȱʡߵpaint.</zh>
   * <en>The default paint used for drawing the guide lines.</en>
   */
  public static final Paint DEFAULT_GUIDE_LINE_PAINT = Color.black ;

  /**
   * <zh>ȱʡߵstroke.</zh>
   * <en>The default stroke used for drawing the guide lines.</en>
   */
  public static final Stroke DEFAULT_GUIDE_LINE_STROKE = new BasicStroke(1.0F) ;

  /**
   * <zh>Ƿɼ.</zh>
   * <en>The boolean flag indicating if guide lines should be drawn for every item.</en>
   */
  private boolean guideLineVisible ;

  /**
   * <zh>ߵpaint.</zh>
   * <en>The paint used for drawing the guide lines.</en>
   */
  private transient Paint guideLinePaint ;

  /**
   * <zh>ߵstroke.</zh>
   * <en>The stroke used for drawing the guide lines.</en>
   */
  private transient Stroke guideLineStroke ;

  /**
   * <zh>ʵֽӿ{@link IPaintScale}Ķ.</zh>
   * <en>The paint scale.</en>
   */
  private IPaintScale paintScale ;

  /**
   * <zh>ͼǩ.</zh>
   * <en>The legend label.</en>
   */
  private String legendLabel ;

  /**
   * <zh>.</zh>
   * <en>The constructor.</en>
   *
   * @param dataset SingleRowDataset <zh>ݼ (Ϊnull).</zh>
   * <en>The dataset (null not permitted).</en>
   */
  public XYShapeRender(IMatrixDataset dataset) {
    this(dataset, DEFAULT_PAINT_SCALE, DEFAULT_LEGEND_LABEL) ;
  }

  /**
   * <zh>.</zh>
   * <en>The constructor.</en>
   *
   * @param dataset SingleRowDataset <zh>ݼ (Ϊnull).</zh>
   * <en>The dataset (null not permitted).</en>
   * @param legendLabel String <zh>ͼǩ.</zh>
   * <en>The legend label.</en>
   */
  public XYShapeRender(IMatrixDataset dataset, String legendLabel) {
    this(dataset, DEFAULT_PAINT_SCALE, legendLabel) ;
  }

  /**
   * <zh>.</zh>
   * <en>The constructor.</en>
   *
   * @param dataset SingleRowDataset <zh>ݼ (Ϊnull).</zh>
   * <en>The dataset (null not permitted).</en>
   * @param paintScale <zh>IPaintScale.</zh>
   * <en>The paint scale.</en>
   * @param legendLabel String <zh>ͼǩ.</zh>
   * <en>The legend label.</en>
   */
  public XYShapeRender(IMatrixDataset dataset, IPaintScale paintScale, String legendLabel) {
    super(dataset) ;
    if (paintScale == null) {
      throw new IllegalArgumentException("Null 'paintScale' argument") ;
    }

    this.paintScale = paintScale ;
    this.legendLabel = legendLabel ;

    this.guideLineVisible = DEFAULT_GUIDE_LINE_VISIBLE ;
    this.guideLinePaint = DEFAULT_GUIDE_LINE_PAINT ;
    this.guideLineStroke = DEFAULT_GUIDE_LINE_STROKE ;
  }

  /**
   * <zh>Ƿɼ.</zh>
   * <en>Returns the flag indicating if guide lines should be drawn for every item.</en>
   *
   * @return boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public boolean isGuideLineVisible() {
    return this.guideLineVisible ;
  }

  /**
   * <zh>Ƿɼ,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the flag indicating if guide lines should be drawn for every item and sends a
   * {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param flag boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public void setGuideLineVisible(boolean flag) {
    if (this.guideLineVisible != flag) {
      this.guideLineVisible = flag ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>ػߵpaint.</zh>
   * <en>Returns the paint used for drawing the guide lines.</en>
   *
   * @return Paint <zh>paint (Ϊnull).</zh>
   * <en>The paint (never null).</en>
   */
  public Paint getGuideLinePaint() {
    return this.guideLinePaint ;
  }

  /**
   * <zh>ûߵpaint,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the paint used for drawing the guide lines and sends a
   * {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param paint Paint <zh>ߵpaint (Ϊnull).</zh>
   * <en>The paint used for drawing the guide lines (null not permitted).</en>
   */
  public void setGuideLinePaint(Paint paint) {
    if (paint == null) {
      throw new IllegalArgumentException("Null 'paint' argument.") ;
    }

    this.guideLinePaint = paint ;
    fireRenderChanged() ;
  }

  /**
   * <zh>ػߵstroke.</zh>
   * <en>Returns the stroke used for drawing the guide lines.</en>
   *
   * @return Stroke <zh>stroke (Ϊnull).</zh>
   * <en>The stroke (never null).</en>
   */
  public Stroke getGuideLineStroke() {
    return this.guideLineStroke ;
  }

  /**
   * <zh>ûߵstroke,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the stroke used for drawing the guide lines and sends a
   * {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param stroke Stroke <zh>ߵstroke (Ϊnull).</zh>
   * <en>The stroke used for drawing the guide lines (null not permitted).</en>
   */
  public void setGuideLineStroke(Stroke stroke) {
    if (stroke == null) {
      throw new IllegalArgumentException("Null 'stroke' argument.") ;
    }

    this.guideLineStroke = stroke ;
    fireRenderChanged() ;
  }

  /**
   * <zh>IPaintScaleĶ.</zh>
   * <en>Returns the paint scale.</en>
   *
   * @return IPaintScale <zh>IPaintScaleĶ.</zh>
   * <en>The paint scale.</en>
   */
  public IPaintScale getPaintScale() {
    return this.paintScale ;
  }

  /**
   * <zh>IPaintScaleĶ,м{@link com.ijchart.xychart.event.RenderChangeEvent}
   * ¼.</zh>
   * <en>Sets the paint scale and sends a
   * {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param paintScale IPaintScale <zh>IPaintScaleĶ.</zh>
   * <en>The paint scale.</en>
   */
  public void setPaintScale(IPaintScale paintScale) {
    if (paintScale == null) {
      throw new IllegalArgumentException("Null 'paintScale' argument") ;
    }

    this.paintScale = paintScale ;
    fireRenderChanged() ;
  }

  /**
   * <zh>ͼǩ.</zh>
   * <en>Returns the legend label.</en>
   *
   * @return String <zh>ͼǩ (Ϊnull).</zh>
   * <en>The legend label (maybe null).</en>
   */
  public String getLegendLabel() {
    return this.legendLabel ;
  }

  /**
   * <zh>ͼǩ.</zh>
   * <en>The legend label.</en>
   *
   * @param label String <zh>ͼǩ (Ϊnull).</zh>
   * <en>The legend label (null permitted).</en>
   */
  public void setLegendLabel(String label) {
    if (EqualsUtil.equals(this.legendLabel, label)) {
      this.legendLabel = label ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>еͼ.</zh>
   * <en>Draws all data items.</en>
   *
   * @param xAxis AbstractValueAxis <zh>xֵ (Ϊnull).</zh>
   * <en>The x value axis (null not permitted).</en>
   * @param xAxisEdge RectangleEdge <zh>xλ(Ϊnull).</zh>
   * <en>The x axis edge.</en>
   * @param yAxis AbstractValueAxis <zh>yֵ (Ϊnull).</zh>
   * <en>The y value axis (null not permitted).</en>
   * @param yAxisEdge RectangleEdge <zh>yλ (Ϊnull).</zh>
   * <en>The y axis edge (null not permitted).</en>
   * @param dataArea Rectangle2D <zh> (Ϊnull).</zh>
   * <en>The chart data area (null not permitted).</en>
   * @param orientation ChartOrientation <zh>ͼķ (Ϊnull).</zh>
   * <en>The chart orientation (null not permitted).</en>
   * @param renderShapes RenderShapeCollection <zh>renderͼμ (Ϊnull).</zh>
   * <en>Collects information about the render (null not permitted).</en>
   * @param alpha float <zh>alphaֵ,ڱ[0.0,1.0].</zh>
   * <en>The alpha transparency (in the range 0.0f to 1.0f, where 0.0f is fully transparent,
   * and 1.0f is fully opaque).</en>
   */
  public void drawAllItems(AbstractValueAxis xAxis, RectangleEdge xAxisEdge,
                           AbstractValueAxis yAxis, RectangleEdge yAxisEdge,
                           Rectangle2D dataArea, ChartOrientation orientation,
                           RenderShapeCollection renderShapes, float alpha) {
    if (xAxis == null) {
      throw new IllegalArgumentException("Null 'xAxis' argument.") ;
    }
    if (xAxisEdge == null) {
      throw new IllegalArgumentException("Null 'xAxisEdge' argument.") ;
    }
    if (yAxis == null) {
      throw new IllegalArgumentException("Null 'yAxis' argument.") ;
    }
    if (yAxisEdge == null) {
      throw new IllegalArgumentException("Null 'yAxisEdge' argument.") ;
    }
    if (dataArea == null) {
      throw new IllegalArgumentException("Null 'dataArea' argument.") ;
    }
    if (orientation == null) {
      throw new IllegalArgumentException("Null 'orientation' argument.") ;
    }
    if (renderShapes == null) {
      throw new IllegalArgumentException("Null 'renderShapes' argument.") ;
    }
    if (alpha < 0.0f || alpha > 1.0f) {
      throw new IllegalArgumentException("The 'alpha' value must be in the range 0.0f to 1.0f.") ;
    }

    IMatrixDataset dataset = getDataset() ;
    if (dataset.isEmpty()) {
      return ;
    }

    Shape shape = getDrawingSupplier().getNextShape() ;
    Rectangle2D shapeBound = shape.getBounds2D() ;
    int rows = dataset.getRowCount() ;

    getFillPaints(rows) ;
    getBorderPaints(rows) ;
    this.borderStrokes = getBorderStrokes(rows) ;
    boolean isVertical = orientation.equals(ChartOrientation.VERTICAL) ;

    for (int i = 0 ; i < rows ; i++) {
      int currRowCols = dataset.getColumnCount(i) ;
      for (int j = 0 ; j < currRowCols ; j++) {
        IChartData data = dataset.getData(i, j) ;
        if (data != null && data instanceof XYZNumberChartData) {
          XYZNumberChartData xyz = (XYZNumberChartData) data ;
          double x = xyz.getX() ;
          double y = xyz.getY() ;
          double z = xyz.getZ() ;
          if (Double.isNaN(x) || Double.isNaN(y) || Double.isNaN(z)) {
            renderShapes.addDataItemShape(null) ;
          } else {
            double xPos = xAxis.valueToJava2D(dataArea, xAxisEdge, x, false) ;
            double yPos = yAxis.valueToJava2D(dataArea, yAxisEdge, y, false) ;
            if (xPos == AbstractValueAxis.INVALID_COORDINATE_VALUE
                || yPos == AbstractValueAxis.INVALID_COORDINATE_VALUE) {
              renderShapes.addDataItemShape(null) ;
              continue ;
            }

            Paint p = this.paintScale.getPaint(z) ;
            LineAndShapeChartShape tmpShape = null ;
            if (isVertical) {
              tmpShape = new LineAndShapeChartShape(xPos, yPos, shape, null) ;
            } else {
              tmpShape = new LineAndShapeChartShape(yPos, xPos, shape, null) ;
            }

            IChartShape chartShape = createItemShape(orientation, yAxis, i, j, shapeBound, p,
                                                     p, this.borderStrokes[i], alpha, tmpShape) ;
            renderShapes.addDataItemShape(chartShape) ;

            if (this.guideLineVisible) {
              boolean[] result = hasXYGuidLine(x, y, i, j) ;
              if (!result[0]) { //draw x guid line
                LineChartShape lineShape = null ;
                if (isVertical) {
                  lineShape = new LineChartShape(xPos, dataArea.getMinY(), xPos, dataArea.getMaxY()) ;
                } else {
                  lineShape = new LineChartShape(dataArea.getMinX(), xPos, dataArea.getMaxX(), xPos) ;
                }

                lineShape.setBorderPaint(this.guideLinePaint) ;
                lineShape.setBorderStroke(this.guideLineStroke) ;
                renderShapes.addGuideLineShape(lineShape) ;
              }

              if (!result[1]) { //draw y guid line
                LineChartShape lineShape = null ;
                if (isVertical) {
                  lineShape = new LineChartShape(dataArea.getMinX(), yPos, dataArea.getMaxX(), yPos) ;
                } else {
                  lineShape = new LineChartShape(yPos, dataArea.getMinY(), yPos, dataArea.getMaxY()) ;
                }

                lineShape.setBorderPaint(this.guideLinePaint) ;
                lineShape.setBorderStroke(this.guideLineStroke) ;
                renderShapes.addGuideLineShape(lineShape) ;
              }
            }
          }
        } else {
          renderShapes.addDataItemShape(null) ;
        }
      }
    }
  }

  /**
   * <zh>жǷҪx,y.</zh>
   * <en>Whether or not draw guid line for specified data.</en>
   *
   * @param x double
   * @param y double
   * @param row int
   * @param column int
   *
   * @return boolean[]
   */
  private boolean[] hasXYGuidLine(double x, double y, int row, int column) {
    boolean[] result = new boolean[2] ;

    IMatrixDataset dataset = getDataset() ;
    int rows = dataset.getRowCount() ;

    for (int i = 0 ; i < rows ; i++) {
      int currRowCols = dataset.getColumnCount(i) ;
      for (int j = 0 ; j < currRowCols ; j++) {
        if (i == row && j >= column) {
          return result ;
        }

        IChartData data = dataset.getData(i, j) ;
        if (data != null && data instanceof XYZNumberChartData) {
          XYZNumberChartData xyz = (XYZNumberChartData) data ;

          if (x == xyz.getX()) {
            result[0] = true ;
          }
          if (y == xyz.getY()) {
            result[1] = true ;
          }
        }
      }
    }

    return result ;
  }

  /**
   * <zh>ֻͼ,ͼε.</zh>
   * <en>Create a shape with no attribute specified.</en>
   *
   * @param rect Rectangle2D <zh>ͼľ (Ϊnull).</zh>
   * <en>The bar rectangle (null not permitted).</en>
   * @param text String <zh>Ҫʾ (Ϊnull).</zh>
   * <en>The text (null permitted).</en>
   * @param appendParam Object <zh>ӵĲ,ڴݶĶ (Ϊnull).</zh>
   * <en>The append parameter,used for special propose (null permitted).</en>
   *
   * @return IChartShape <zh>ͼ.</zh>
   * <en>The shape.</en>
   */
  protected IChartShape createItemShapeOnly(Rectangle2D rect, String text,
                                            Object appendParam) {
    if (appendParam == null) {
      throw new IllegalArgumentException("Null 'appendParam' argument.") ;
    }

    if (appendParam instanceof LineAndShapeChartShape) {
      LineAndShapeChartShape lineAndShape = (LineAndShapeChartShape) appendParam ;
      lineAndShape.setText(text) ;
    }

    return (IChartShape) appendParam ;
  }

  /**
   * <zh>ͼĸ.</zh>
   * <en>Returns the count of legends.</en>
   *
   * @return int <zh>ͼĸ.</zh>
   * <en>The count of legends.</en>
   */
  protected int getLegendItemCount() {
    return 1 ;
  }

  /**
   * <zh>ͼͼεıǩ.</zh>
   * <en>Returns the label array for legend item shapes.</en>
   *
   * @return String[] <zh>ǩ</zh>
   * <en>The label array.</en>
   */
  protected String[] getLegendItemLabels() {
    return new String[] {""} ;
  }

  /**
   * <zh>ÿͼͼ.</zh>
   * <en>Returns the shape array of legend items.</en>
   *
   * @param legendCount int <zh>ͼĸ.</zh>
   * <en>The legend items count.</en>
   * @return IChartShape[] <zh>ͼ.</zh>
   * <en>The shapes array.</en>
   */
  protected IChartShape[] getLegendItemShapes(int legendCount) {
    if (legendCount > 0) {
      PaintScaleLegendChartShape shape =
        new PaintScaleLegendChartShape(this.paintScale, this.legendLabel,
                                       getLegendPositon(), getChartArea(), getDataset()) ;
      IChartShape[] shapes = new IChartShape[legendCount] ;
      for (int i = 0 ; i < legendCount ; i++) {
        if (i == 0) {
          shapes[i] = shape ;
        } else {
          shapes[i] = null ;
        }
      }

      return shapes ;
    }

    return null ;
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Indicates if some other object is "equal to" this one.</en>
   *
   * @param obj <zh>ҪȽϵĶ</zh>
   * <en>the reference object with which to compare.</en>
   * @return <zh>ȷtrue,򷵻false.</zh>
   * <en>true if this object is the same as the obj argument; false otherwise.</en>
   */
  public boolean equals(Object obj) {
    if (this == obj) {
      return true ;
    }
    if (!(obj instanceof XYShapeRender)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    XYShapeRender that = (XYShapeRender) obj ;
    if (this.guideLineVisible != that.guideLineVisible) {
      return false ;
    }
    if (!EqualsUtil.equals(this.guideLinePaint, that.guideLinePaint)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.guideLineStroke, that.guideLineStroke)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.paintScale, that.paintScale)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.legendLabel, that.legendLabel)) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    double hashcode = super.hashCode() ;
    hashcode = 37 * hashcode + HashcodeUtil.hashCode(this.guideLineVisible) ;
    hashcode = 37 * hashcode + HashcodeUtil.hashCode(this.guideLinePaint) ;
    hashcode = 37 * hashcode + HashcodeUtil.hashCode(this.guideLineStroke) ;
    hashcode = 37 * hashcode + HashcodeUtil.hashCode(this.paintScale) ;
    hashcode = 37 * hashcode + HashcodeUtil.hashCode(this.legendLabel) ;

    return (int) hashcode ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   *   java.lang.Cloneable
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a string representation of the object.</en>
   *
   * @return  a string representation of the object.
   */
  public String toString() {
    return "XYShapeRender" ;
  }

}
